"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.areCompatible = exports.getCompatibility = exports.getNativeAppCompatibilityReport = exports.logCompatibilityReportTable = exports.checkCompatibilityWithPlatform = exports.checkCompatibilityWithNativeApp = void 0;
const ern_core_1 = require("ern-core");
const ern_cauldron_api_1 = require("ern-cauldron-api");
const lodash_1 = __importDefault(require("lodash"));
const chalk_1 = __importDefault(require("chalk"));
const cli_table_1 = __importDefault(require("cli-table"));
const semver_1 = __importDefault(require("semver"));
//
// Check compatibility of a given miniapp against one or multiple native apps
async function checkCompatibilityWithNativeApp(miniApp, appName, platformName, versionName) {
    const compatReport = await ern_core_1.kax.task('Checking compatibility').run(getNativeAppCompatibilityReport(miniApp, {
        appName,
        platformName,
        versionName,
    }));
    for (const r of compatReport) {
        ern_core_1.log.info(chalk_1.default.magenta(`${r.appName}:${r.appPlatform}:${r.appVersion}`) +
            ' : ' +
            (r.isCompatible
                ? chalk_1.default.green('COMPATIBLE')
                : chalk_1.default.red('NOT COMPATIBLE')));
        logCompatibilityReportTable(r.compatibility);
        if (appName && platformName && versionName) {
            return r;
        }
    }
}
exports.checkCompatibilityWithNativeApp = checkCompatibilityWithNativeApp;
//
// Check compatibility of a given miniapp against a given platform version
async function checkCompatibilityWithPlatform(miniApp, platformVersion) {
    const miniappDependencies = miniApp.getPackageJsonDependencies();
    const platformDependencies = await ern_core_1.manifest.getJsAndNativeDependencies({
        platformVersion,
    });
    const report = getCompatibility(miniappDependencies, platformDependencies);
    const isCompatible = report.incompatible.length === 0;
    ern_core_1.log.info(isCompatible ? chalk_1.default.green('COMPATIBLE') : chalk_1.default.red('NOT COMPATIBLE'));
    logCompatibilityReportTable(report);
}
exports.checkCompatibilityWithPlatform = checkCompatibilityWithPlatform;
//
// Log compatiblity report to terminal in a fancy table
async function logCompatibilityReportTable(report) {
    const table = new cli_table_1.default({
        colWidths: [40, 16, 15],
        head: [
            chalk_1.default.cyan('Name'),
            chalk_1.default.cyan('Needed Version'),
            chalk_1.default.cyan('Local Version'),
        ],
    });
    for (const compatibleEntry of report.compatible) {
        table.push([
            compatibleEntry.dependencyName,
            chalk_1.default.green(compatibleEntry.remoteVersion ? compatibleEntry.remoteVersion : ''),
            chalk_1.default.green(compatibleEntry.localVersion ? compatibleEntry.localVersion : ''),
        ]);
    }
    for (const compatibleNonStrictEntry of report.compatibleNonStrict) {
        table.push([
            compatibleNonStrictEntry.dependencyName,
            chalk_1.default.yellow(compatibleNonStrictEntry.remoteVersion
                ? compatibleNonStrictEntry.remoteVersion
                : ''),
            chalk_1.default.yellow(compatibleNonStrictEntry.localVersion
                ? compatibleNonStrictEntry.localVersion
                : ''),
        ]);
    }
    for (const incompatibleEntry of report.incompatible) {
        table.push([
            incompatibleEntry.dependencyName,
            incompatibleEntry.remoteVersion,
            chalk_1.default.red(incompatibleEntry.localVersion),
        ]);
    }
    ern_core_1.log.info(table.toString());
}
exports.logCompatibilityReportTable = logCompatibilityReportTable;
//
// Retrieve compatibility report(s) of a given miniapp against one or multiple native apps
async function getNativeAppCompatibilityReport(miniApp, { appName, platformName, versionName, }) {
    const result = [];
    const cauldron = await (0, ern_cauldron_api_1.getActiveCauldron)();
    const nativeApps = await cauldron.getAllNativeApps();
    // Todo : pass miniapp to these functions instead (or just move compat methods in MiniApp class maybe)
    const nativeDependencies = await miniApp.getNativeDependencies();
    const miniappDependencies = [
        ...nativeDependencies.apis,
        ...nativeDependencies.nativeApisImpl,
        ...nativeDependencies.thirdPartyInManifest,
    ];
    // I so love building pyramids !!! :P
    for (const nativeApp of nativeApps) {
        if (!appName || nativeApp.name === appName) {
            for (const nativeAppPlatform of nativeApp.platforms) {
                if (!platformName || nativeAppPlatform.name === platformName) {
                    for (const nativeAppVersion of nativeAppPlatform.versions) {
                        if (!versionName || nativeAppVersion.name === versionName) {
                            const napDescriptor = new ern_core_1.AppVersionDescriptor(nativeApp.name, nativeAppPlatform.name, nativeAppVersion.name);
                            const nativeAppDependencies = await cauldron.getNativeDependencies(napDescriptor);
                            const compatibility = getCompatibility(miniappDependencies, nativeAppDependencies, {
                                uncompatibleIfARemoteDepIsMissing: nativeAppVersion.isReleased,
                            });
                            result.push({
                                appBinary: nativeAppVersion.binary,
                                appName: nativeApp.name,
                                appPlatform: nativeAppPlatform.name,
                                appVersion: nativeAppVersion.name,
                                compatibility,
                                isCompatible: compatibility.incompatible.length === 0,
                                isReleased: nativeAppVersion.isReleased,
                            });
                        }
                    }
                }
            }
        }
    }
    return result;
}
exports.getNativeAppCompatibilityReport = getNativeAppCompatibilityReport;
//
// Get a compatibility object containing the compatibility result
// from comparing an array of dependencies (localDeps) against anoter
// array of dependencies (remoteDeps)
//
// Input : localDeps/remoteDeps sample array :
//
// [{
//   name: "react-native-electrode-bridge",
//   scope: "walmart",
//   version: "1.0.0"
// }]
//
// Output object example :
//
// {
//   compatible: [
//     {
//       dependencyName: "react-native-electrode-bridge",
//       scope: "walmart",
//       localVersion: "1.0.0",
//       remoteVersion: "1.0.0"
//     }
//   ],
//   compatibleNonStrict: [
//     {
//       dependencyName: "react-native-cookie-api",
//       scope: "walmart",
//       localVersion: "1.0.0",
//       remoteVersion: "1.3.0"
//     }
//   ],
//   incompatible: [
//     {
//       dependencyName: "react-native",
//       localVersion: "0.38.2",
//       remoteVersion: "0.40.0"
//     }
//   ]
// }
//
// Optional inputs :
// - uncompatibleIfARemoteDepIsMissing : true if a missing remote
// dependency should be deemed uncompatible (for example for released
// native application versions, if a native dependency is missing from
// the binary, it's not compatible. On the other hand, for non released
// versions it's OK as it's always possible to regenerate a container
// that include this new native dependency)
function getCompatibility(localDeps, remoteDeps, { uncompatibleIfARemoteDepIsMissing, } = {}) {
    const result = {
        compatible: [],
        compatibleNonStrict: [],
        incompatible: [],
    };
    for (const remoteDep of remoteDeps) {
        const localDep = lodash_1.default.find(localDeps, (d) => remoteDep.same(d, { ignoreVersion: true }));
        const localDepVersion = localDep ? localDep.version : undefined;
        const entry = {
            dependencyName: remoteDep.name,
            localVersion: localDepVersion,
            remoteVersion: remoteDep.version,
        };
        // If a remote dependency exists locally the following applies in term
        // of compatibility
        // If same exact version : COMPATIBLE
        // If different version :
        //    - If API or BRIDGE :
        //      + If local version < remote version but MAJOR is same :
        //          => COMPATIBLE (backward compatibility of APIs / Bridge)
        //      + If local version < remove version but MAJOR is different :
        //          => INCOMPATIBLE (breaking change)
        //      + If local version > remote version
        //          => INCOMPATIBLE (could be using feature not present in older)
        //    - If third party plugin :
        //      => INCOMPATIBLE
        if (localDep && localDepVersion && localDepVersion !== remoteDep.version) {
            // Todo : do not infer api or api-impl by looking solely at the suffix as its not mandatory anymore, so
            // we might get false negatives
            if (localDep.name.endsWith('-api') ||
                localDep.name.endsWith('-api-impl') ||
                localDep.name === 'react-native-electrode-bridge') {
                if (semver_1.default.major(localDepVersion) ===
                    semver_1.default.major(remoteDep.version)) {
                    if (semver_1.default.lt(localDepVersion, remoteDep.version)) {
                        result.compatibleNonStrict.push(entry);
                    }
                    else {
                        result.incompatible.push(entry);
                    }
                }
                else {
                    result.incompatible.push(entry);
                }
            }
            else {
                result.incompatible.push(entry);
            }
        }
        else if (localDepVersion && localDepVersion === remoteDep.version) {
            result.compatible.push(entry);
        }
    }
    if (uncompatibleIfARemoteDepIsMissing) {
        for (const localDep of localDeps) {
            const remoteDep = lodash_1.default.find(remoteDeps, (d) => d.name === localDep.name);
            const remoteDepVersion = remoteDep ? remoteDep.version : undefined;
            const entry = {
                dependencyName: localDep.name,
                localVersion: localDep.version,
                remoteVersion: remoteDepVersion || 'MISSING',
            };
            if (!remoteDepVersion) {
                result.incompatible.push(entry);
            }
        }
    }
    return result;
}
exports.getCompatibility = getCompatibility;
async function areCompatible(miniApps, targetDescriptor) {
    for (const miniApp of miniApps) {
        const miniAppInstance = await ern_core_1.kax
            .task(`Checking native dependencies version alignment of ${miniApp} with ${targetDescriptor}`)
            .run(ern_core_1.MiniApp.fromPackagePath(new ern_core_1.PackagePath(miniApp.toString())));
        const report = await checkCompatibilityWithNativeApp(miniAppInstance, targetDescriptor.name, targetDescriptor.platform || undefined, targetDescriptor.version || undefined);
        if (!report.isCompatible) {
            ern_core_1.log.warn('At least one native dependency version is not aligned !');
            return false;
        }
        else {
            ern_core_1.log.info(`${miniApp} native dependencies versions are aligned with ${targetDescriptor}`);
        }
    }
    return true;
}
exports.areCompatible = areCompatible;
//# sourceMappingURL=compatibility.js.map