"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.alignPackageJsonOnManifest = exports.deleteBranchOrTag = exports.deleteTag = exports.deleteBranch = exports.createBranchOrTag = exports.createTag = exports.createBranch = exports.getGitHubApi = void 0;
const ern_core_1 = require("ern-core");
const ern_cauldron_api_1 = require("ern-cauldron-api");
async function getGitHubApi({ owner, repo, }) {
    if (!process.env.ERN_GITHUB_TOKEN) {
        throw new Error('ERN_GITHUB_TOKEN environment variable must be set');
    }
    const cauldron = await (0, ern_cauldron_api_1.getActiveCauldron)();
    const gitHubConfig = await cauldron.getGitHubConfig();
    const opts = {
        auth: `token ${process.env.ERN_GITHUB_TOKEN}`,
    };
    if (gitHubConfig && gitHubConfig.baseUrl) {
        opts.baseUrl = gitHubConfig.baseUrl;
    }
    return new ern_core_1.GitHubApi({
        opts,
        owner,
        repo,
    });
}
exports.getGitHubApi = getGitHubApi;
async function createBranch({ name, packages, }) {
    return createBranchOrTag({ name, packages, type: 'branch' });
}
exports.createBranch = createBranch;
async function createTag({ name, packages, }) {
    return createBranchOrTag({ name, packages, type: 'tag' });
}
exports.createTag = createTag;
async function createBranchOrTag({ name, packages, type, }) {
    for (const pkg of packages) {
        if (!pkg.isGitPath) {
            ern_core_1.log.info(`Skipping ${pkg.basePath} [not git based]`);
            continue;
        }
        const opts = { name };
        const { owner, repo } = extractGitData(pkg);
        const api = await getGitHubApi({ owner, repo });
        const refExist = type === 'branch' ? await api.isBranch(name) : await api.isTag(name);
        if (refExist) {
            ern_core_1.log.warn(`Skipping ${pkg.basePath} [${name} ${type} already exist]`);
            continue;
        }
        if (pkg.version) {
            if (await api.isSha(pkg.version)) {
                opts.fromSha = pkg.version;
            }
            else if (await api.isBranch(pkg.version)) {
                opts.fromBranch = pkg.version;
            }
            else if (await api.isTag(pkg.version)) {
                opts.fromTag = pkg.version;
            }
            else {
                throw new Error(`${pkg.version} is not a valid git ref of ${pkg.basePath}`);
            }
        }
        if (type === 'branch') {
            await api.createBranch(opts);
        }
        else {
            await api.createTag(opts);
        }
        ern_core_1.log.info(`Created ${type} ${name} from ${pkg.version} in ${pkg.basePath}`);
    }
}
exports.createBranchOrTag = createBranchOrTag;
async function deleteBranch({ name, packages, }) {
    return deleteBranchOrTag({ name, packages, type: 'branch' });
}
exports.deleteBranch = deleteBranch;
async function deleteTag({ name, packages, }) {
    return deleteBranchOrTag({ name, packages, type: 'tag' });
}
exports.deleteTag = deleteTag;
async function deleteBranchOrTag({ name, packages, type, }) {
    for (const pkg of packages) {
        if (!pkg.isGitPath) {
            ern_core_1.log.info(`Skipping ${pkg.basePath} [not git based]`);
            continue;
        }
        const opts = { name };
        const { owner, repo } = extractGitData(pkg);
        const api = await getGitHubApi({ owner, repo });
        const refExist = type === 'branch' ? await api.isBranch(name) : await api.isTag(name);
        if (!refExist) {
            ern_core_1.log.warn(`Skipping ${pkg.basePath} [${name} ${type} as it does not exist]`);
            continue;
        }
        if (type === 'branch') {
            await api.deleteBranch(opts);
        }
        else {
            await api.deleteTag(opts);
        }
        ern_core_1.log.info(`Deleted ${type} ${name} from ${pkg.basePath}`);
    }
}
exports.deleteBranchOrTag = deleteBranchOrTag;
async function alignPackageJsonOnManifest({ manifestId, packages, }) {
    for (const pkg of packages) {
        if (!pkg.isGitPath) {
            ern_core_1.log.debug(`Skipping ${pkg.basePath} [not git based]`);
            continue;
        }
        ern_core_1.log.info(`Aligning dependencies of ${pkg.basePath}`);
        const { owner, repo } = extractGitData(pkg);
        const gitHubApi = await getGitHubApi({
            owner,
            repo,
        });
        const res = await gitHubApi.getFileContent({
            fromBranch: pkg.version,
            path: 'package.json',
        });
        const jsonRes = JSON.parse(res);
        const wasUpdated = await updatePackageJson({
            manifestId,
            packageJson: jsonRes,
        });
        if (!wasUpdated) {
            ern_core_1.log.info(`All dependencies of ${pkg.toString()} are already aligned. Skipping.`);
        }
        else {
            const newContent = JSON.stringify(jsonRes, null, 2);
            await gitHubApi.updateFileContent({
                commitMessage: `Align dependencies on ${manifestId} manifest id`,
                newContent,
                onBranch: pkg.version,
                path: 'package.json',
            });
            ern_core_1.log.info(`Successfully aligned dependencies of ${pkg.basePath}.`);
        }
    }
}
exports.alignPackageJsonOnManifest = alignPackageJsonOnManifest;
async function updatePackageJson({ manifestId, packageJson, }) {
    let wasUpdated = false;
    const manifestDependencies = await ern_core_1.manifest.getJsAndNativeDependencies({
        manifestId,
    });
    for (const manifestDependency of manifestDependencies) {
        if (packageJson.dependencies[manifestDependency.basePath]) {
            const dependencyManifestVersion = manifestDependency.version;
            const localDependencyVersion = packageJson.dependencies[manifestDependency.basePath];
            if (dependencyManifestVersion !== localDependencyVersion) {
                ern_core_1.log.info(`${manifestDependency.basePath} : ${localDependencyVersion} => ${dependencyManifestVersion}`);
                packageJson.dependencies[manifestDependency.basePath] =
                    dependencyManifestVersion;
                wasUpdated = true;
            }
        }
    }
    return wasUpdated;
}
function extractGitData(p) {
    if (!p.isGitPath) {
        throw new Error('Not a git package path !');
    }
    const [, owner, repo] = p.basePath.startsWith('https')
        ? /^https:\/\/[^\/]+\/([^\/]+)\/(.+).git$/.exec(p.basePath)
        : /^git\+ssh:\/\/[^\/]+\/([^\/]+)\/(.+).git$/.exec(p.basePath);
    return {
        owner,
        repo,
        version: p.version,
    };
}
//# sourceMappingURL=gitHub.js.map