"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ern_core_1 = require("ern-core");
const ern_cauldron_api_1 = require("ern-cauldron-api");
const ern_composite_gen_1 = require("ern-composite-gen");
const lodash_1 = __importDefault(require("lodash"));
const path_1 = __importDefault(require("path"));
const fs_extra_1 = __importDefault(require("fs-extra"));
const os_1 = __importDefault(require("os"));
async function start({ baseComposite, compositeDir, jsApiImpls, miniapps = [], descriptor, flavor, launchArgs, launchEnvVars, launchFlags, watchNodeModules = [], packageName, activityName, bundleId, extraJsDependencies, disableBinaryStore, host, port, resetCache, } = {}) {
    const cauldron = await (0, ern_cauldron_api_1.getActiveCauldron)({ throwIfNoActiveCauldron: false });
    if (!cauldron && descriptor) {
        throw new Error('To use a native application descriptor, a Cauldron must be active');
    }
    if (!miniapps && !jsApiImpls && !descriptor) {
        throw new Error('Either miniapps, jsApiImpls or descriptor needs to be provided');
    }
    miniapps
        .filter((m) => m.isFilePath)
        .forEach((m) => {
        ern_core_1.kax.task(`Auto linking local MiniApp from ${m.basePath}`).succeed();
    });
    let resolutions;
    if (descriptor && cauldron) {
        miniapps = await cauldron.getContainerMiniApps(descriptor, {
            favorGitBranches: true,
        });
        const compositeGenConfig = await cauldron.getCompositeGeneratorConfig(descriptor);
        baseComposite =
            baseComposite !== null && baseComposite !== void 0 ? baseComposite : ((compositeGenConfig === null || compositeGenConfig === void 0 ? void 0 : compositeGenConfig.baseComposite) &&
                ern_core_1.PackagePath.fromString(compositeGenConfig.baseComposite));
        resolutions = compositeGenConfig === null || compositeGenConfig === void 0 ? void 0 : compositeGenConfig.resolutions;
    }
    compositeDir = compositeDir || (0, ern_core_1.createTmpDir)();
    ern_core_1.log.trace(`Temporary composite directory is ${compositeDir}`);
    const miniAppsLinksObj = ern_core_1.config.get('miniAppsLinks', {});
    const linkedMiniAppsPkgNames = Object.keys(miniAppsLinksObj).filter((p) => fs_extra_1.default.existsSync(miniAppsLinksObj[p]));
    //
    // Replace all npm provided miniapps that have a linked directory,
    // with their linked directory
    const npmMiniAppsPkgNames = miniapps
        .filter((m) => m.isRegistryPath)
        .map((m) => m.name);
    const npmMiniAppsLinkedPkgNames = lodash_1.default.intersection(linkedMiniAppsPkgNames, npmMiniAppsPkgNames);
    lodash_1.default.remove(miniapps, (m) => npmMiniAppsLinkedPkgNames.includes(m.name));
    npmMiniAppsLinkedPkgNames.forEach((pkgName) => miniapps.push(ern_core_1.PackagePath.fromString(miniAppsLinksObj[pkgName])));
    const composite = await ern_core_1.kax.task('Generating MiniApps composite').run(ern_composite_gen_1.GeneratedComposite.generate({
        baseComposite,
        extraJsDependencies: extraJsDependencies || undefined,
        jsApiImplDependencies: jsApiImpls,
        miniApps: miniapps,
        outDir: compositeDir,
        resolutions,
    }));
    const compositeMiniApps = await composite.getMiniAppsPackages();
    // Only case we need to account for, is any miniapps passed to
    // start as git remotes which have a local linked directory
    const gitMiniApps = compositeMiniApps.filter((m) => m.packagePath.isGitPath);
    const gitMiniAppsPkgNames = gitMiniApps.map((r) => r.name);
    const gitMiniAppsLinkedPkgNames = lodash_1.default.intersection(linkedMiniAppsPkgNames, gitMiniAppsPkgNames);
    const activeLinkedPkgPaths = gitMiniAppsLinkedPkgNames.map((m) => miniAppsLinksObj[m]);
    if (gitMiniAppsLinkedPkgNames.length > 0) {
        const linkedMiniAppDeps = await (0, ern_core_1.findNativeDependencies)(activeLinkedPkgPaths.map((p) => path_1.default.join(p, 'node_modules')));
        const allNativeDeps = await composite.getNativeDependencies();
        // Exclude api/api impls as they are not native modules
        allNativeDeps.apis = [];
        allNativeDeps.nativeApisImpl = [];
        // Add the ones from composite
        allNativeDeps.thirdPartyInManifest.push(...linkedMiniAppDeps.thirdPartyInManifest);
        allNativeDeps.thirdPartyNotInManifest.push(...linkedMiniAppDeps.thirdPartyNotInManifest);
        allNativeDeps.all.push(...linkedMiniAppDeps.all);
        const allNativeModules = [
            ...allNativeDeps.thirdPartyInManifest,
            ...allNativeDeps.thirdPartyNotInManifest,
        ];
        const dedupedNativeModules = ern_core_1.nativeDepenciesVersionResolution.resolveNativeDependenciesVersionsEx(allNativeDeps);
        const extraNodeModules = {};
        dedupedNativeModules.resolved.forEach((m) => {
            extraNodeModules[m.name] = m.basePath;
        });
        const blacklistRe = lodash_1.default.difference(allNativeModules.map((d) => d.basePath), dedupedNativeModules.resolved.map((d) => d.basePath)).map((l) => new RegExp(os_1.default.platform() === 'win32'
            ? `${l}\\.*`.replace(/\\/g, '\\\\')
            : `${l}\/.*`));
        const compositeLocalMiniappsPaths = compositeMiniApps
            .filter((m) => m.packagePath.isFilePath)
            .map((m) => m.path);
        const allLocalMiniAppsPaths = [
            ...compositeLocalMiniappsPaths,
            ...activeLinkedPkgPaths,
        ];
        const nonInstalledMiniAppsPath = [];
        allLocalMiniAppsPaths.forEach((p) => {
            if (!fs_extra_1.default.pathExistsSync(path_1.default.join(p, 'node_modules'))) {
                nonInstalledMiniAppsPath.push(p);
            }
        });
        if (nonInstalledMiniAppsPath.length > 0) {
            throw new Error(`Some linked MiniApp(s) have not been installed.
Please run 'yarn install' or 'npm install' from the following directories before running 'ern start' :
${nonInstalledMiniAppsPath.join('\n')}
      `);
        }
        await (0, ern_composite_gen_1.patchCompositeBabelRcRoots)({
            cwd: composite.path,
            extraPaths: allLocalMiniAppsPaths,
        });
        await (0, ern_composite_gen_1.createMetroConfig)({
            blacklistRe,
            cwd: composite.path,
            extraNodeModules,
            reactNativeVersion: composite.getReactNativeVersion(),
            watchFolders: allLocalMiniAppsPaths,
        });
        allLocalMiniAppsPaths.forEach((m) => ern_core_1.log.info(`Watching for changes in linked directory ${m}`));
    }
    ern_core_1.reactnative.startPackager({
        cwd: compositeDir,
        host,
        port,
        resetCache,
    });
    if (cauldron && descriptor && !disableBinaryStore) {
        const binaryStoreConfig = await cauldron.getBinaryStoreConfig();
        if (binaryStoreConfig) {
            const cauldronStartCommandConfig = await cauldron.getStartCommandConfig(descriptor);
            const binaryStore = new ern_core_1.ErnBinaryStore(binaryStoreConfig);
            if (await binaryStore.hasBinary(descriptor, { flavor })) {
                if (descriptor.platform === 'android') {
                    if (cauldronStartCommandConfig === null || cauldronStartCommandConfig === void 0 ? void 0 : cauldronStartCommandConfig.android) {
                        packageName =
                            packageName !== null && packageName !== void 0 ? packageName : cauldronStartCommandConfig.android.packageName;
                        activityName =
                            activityName !== null && activityName !== void 0 ? activityName : cauldronStartCommandConfig.android.activityName;
                    }
                    if (!packageName) {
                        throw new Error('You need to provide an Android package name or set it in Cauldron configuration');
                    }
                    const apkPath = await ern_core_1.kax
                        .task('Downloading binary from store')
                        .run(binaryStore.getBinary(descriptor, { flavor }));
                    await ern_core_1.android.runAndroidApk({
                        activityName,
                        apkPath,
                        launchFlags,
                        packageName,
                    });
                }
                else if (descriptor.platform === 'ios') {
                    if (cauldronStartCommandConfig === null || cauldronStartCommandConfig === void 0 ? void 0 : cauldronStartCommandConfig.ios) {
                        bundleId = bundleId !== null && bundleId !== void 0 ? bundleId : cauldronStartCommandConfig.ios.bundleId;
                    }
                    if (!bundleId) {
                        throw new Error('You need to provide an iOS bundle ID or set it in Cauldron configuration');
                    }
                    const appPath = await ern_core_1.kax
                        .task('Downloading binary from store')
                        .run(binaryStore.getBinary(descriptor, { flavor }));
                    await ern_core_1.ios.runIosApp({ appPath, bundleId, launchArgs, launchEnvVars });
                }
            }
        }
    }
}
exports.default = start;
//# sourceMappingURL=start.js.map