import {
  android,
  AppVersionDescriptor,
  bugsnagUpload,
  CodePushPackage,
  createTmpDir,
  getCodePushSdk,
  HermesCli,
  kax,
  log,
  PackagePath,
  reactnative,
  SourceMapStoreSdk,
} from 'ern-core';
import { GeneratedComposite } from 'ern-composite-gen';
import { CauldronNativeApp, getActiveCauldron } from 'ern-cauldron-api';
import * as compatibility from './compatibility';
import _ from 'lodash';
import path from 'path';
import shell from 'shelljs';
import fs from 'fs-extra';
import semver from 'semver';

export async function performCodePushPatch(
  napDescriptor: AppVersionDescriptor,
  deploymentName: string,
  label: string,
  {
    description,
    isDisabled,
    isMandatory,
    rollout,
  }: {
    description?: string;
    isDisabled?: boolean;
    isMandatory?: boolean;
    rollout?: number;
  } = {},
) {
  let cauldron;
  try {
    cauldron = await getActiveCauldron();
    await cauldron.beginTransaction();
    const appName = await getCodePushAppName(napDescriptor);
    const codePushSdk = getCodePushSdk();
    description = description || '';
    await codePushSdk.patch(appName, deploymentName, {
      description,
      isDisabled,
      isMandatory,
      label,
      rollout,
    });
    await cauldron.updateCodePushEntry(napDescriptor, {
      deploymentName,
      description,
      isDisabled,
      isMandatory,
      label,
      rollout,
    });
    await cauldron.commitTransaction(
      `CodePush patched ${napDescriptor.toString()} ${deploymentName} ${label}`,
    );
  } catch (e) {
    if (cauldron) {
      await cauldron.discardTransaction();
    }
    log.error(`[performCodePushPatch] ${e}`);
    throw e;
  }
}

export async function performCodePushPromote(
  sourceNapDescriptor: AppVersionDescriptor,
  targetNapDescriptors: AppVersionDescriptor[],
  sourceDeploymentName: string,
  targetDeploymentName: string,
  {
    description,
    disableDuplicateReleaseError = false,
    force = false,
    label,
    mandatory,
    reuseReleaseBinaryVersion,
    rollout,
    skipNativeDependenciesVersionAlignedCheck = false,
    targetBinaryVersion,
  }: {
    description?: string;
    disableDuplicateReleaseError?: boolean;
    force?: boolean;
    label?: string;
    mandatory?: boolean;
    reuseReleaseBinaryVersion?: boolean;
    rollout?: number;
    skipNativeDependenciesVersionAlignedCheck?: boolean;
    targetBinaryVersion?: string;
  } = {},
) {
  let cauldron;
  try {
    if (reuseReleaseBinaryVersion && targetBinaryVersion) {
      throw new Error(
        `sameBinaryVersionAsRelease and targetBinaryVersion options are mutually exclusive`,
      );
    }

    const codePushSdk = getCodePushSdk();
    cauldron = await getActiveCauldron();
    await cauldron.beginTransaction();
    const cauldronCommitMessage = [
      `CodePush release promotion of ${sourceNapDescriptor.toString()} ${sourceDeploymentName} to ${targetDeploymentName} of`,
    ];

    for (const targetNapDescriptor of targetNapDescriptors) {
      if (!targetNapDescriptor.version) {
        throw new Error(`Missing version in ${targetNapDescriptor.toString()}`);
      }

      const codePushEntrySource = await cauldron.getCodePushEntry(
        sourceNapDescriptor,
        sourceDeploymentName,
        { label },
      );
      if (!codePushEntrySource) {
        throw new Error(
          `No CodePush entry found in Cauldron matching [desc: ${sourceNapDescriptor} dep: ${sourceDeploymentName} label: ${
            label || 'latest'
          }`,
        );
      }

      const miniApps = _.map(codePushEntrySource.miniapps, (miniapp) =>
        PackagePath.fromString(miniapp),
      );
      const jsApiImpls = _.map(codePushEntrySource.jsApiImpls, (jsapiimpl) =>
        PackagePath.fromString(jsapiimpl),
      );

      // If sourceNapDescriptor.version === targetNapDescriptor.version skip the compatibility check
      // As the compatibility check would have already happened as part of the code-push release
      // If skipNativeDependenciesVersionAlignedCheck is set to true
      // and sourceNapDescriptor and targetNapDescriptor versions are different, explicity skip the check
      // would be required for patch version promotion
      if (
        sourceNapDescriptor &&
        sourceNapDescriptor.version !== targetNapDescriptor.version &&
        !skipNativeDependenciesVersionAlignedCheck
      ) {
        const nativeDependenciesVersionAligned =
          await compatibility.areCompatible(miniApps, targetNapDescriptor);

        if (!nativeDependenciesVersionAligned && force) {
          log.warn(
            'Native dependencies versions are not aligned but ignoring due to the use of force flag',
          );
        } else if (!nativeDependenciesVersionAligned && !force) {
          throw new Error(
            'Native dependencies versions of MiniApps are not aligned. Relaunch the operation with the force flag if you wish to ignore.',
          );
        }
      }

      const appName = await getCodePushAppName(sourceNapDescriptor);
      targetBinaryVersion = targetBinaryVersion
        ? targetBinaryVersion
        : reuseReleaseBinaryVersion
        ? codePushEntrySource.metadata.appVersion
        : await buildCodePushTargetBinaryVersion(
            targetNapDescriptor,
            targetDeploymentName,
          );

      description = description || '';
      const codePushResponse: CodePushPackage | void = await kax
        .task(`Promoting release to ${targetBinaryVersion}`)
        .run(
          codePushSdk.promote(
            appName,
            sourceDeploymentName,
            targetDeploymentName,
            {
              appVersion: targetBinaryVersion,
              description,
              isMandatory: !!mandatory,
              label,
              rollout,
            },
            disableDuplicateReleaseError,
          ),
        );

      if (codePushResponse) {
        const sourceYarnLockId = await cauldron.getYarnLockId(
          sourceNapDescriptor,
          sourceDeploymentName,
        );
        if (!sourceYarnLockId) {
          log.error(
            `No yarn.lock was found in source deployment [${sourceDeploymentName} for ${sourceNapDescriptor.toString()}]`,
          );
          log.error(`Skipping promotion to ${targetNapDescriptor.toString()}`);
          continue;
        }
        await cauldron.updateYarnLockId(
          targetNapDescriptor,
          targetDeploymentName,
          sourceYarnLockId,
        );

        await cauldron.addCodePushEntry(
          targetNapDescriptor,
          {
            appVersion: codePushResponse.appVersion,
            deploymentName: targetDeploymentName,
            description: codePushResponse.description,
            isMandatory: codePushResponse.isMandatory,
            label: codePushResponse.label,
            promotedFromLabel: codePushEntrySource.metadata.label,
            releaseMethod: codePushResponse.releaseMethod,
            releasedBy: codePushResponse.releasedBy,
            rollout: codePushResponse.rollout,
            size: codePushResponse.size,
          },
          miniApps,
          jsApiImpls || [],
        );

        // Copy source map if a source map store server is configured
        const sourcemapStoreConfig = await cauldron.getSourceMapStoreConfig();
        if (sourcemapStoreConfig) {
          try {
            const sdk = new SourceMapStoreSdk(sourcemapStoreConfig.url);
            await kax
              .task(
                `Copying source map in source map store [${sourcemapStoreConfig.url}]`,
              )
              .run(
                sdk.copyCodePushSourceMap({
                  deploymentName: sourceDeploymentName,
                  descriptor: sourceNapDescriptor,
                  label: codePushEntrySource.metadata.label!,
                  toDeploymentName: targetDeploymentName,
                  toLabel: codePushResponse.label!,
                  toVersion: targetNapDescriptor.version,
                }),
              );
          } catch (e) {
            log.error(`Source map copy failed : ${e}`);
          }
        }

        cauldronCommitMessage.push(`- ${targetNapDescriptor.toString()}`);
      }
      await kax
        .task('Updating Cauldron')
        .run(cauldron.commitTransaction(cauldronCommitMessage));
    }
  } catch (e) {
    if (cauldron) {
      await cauldron.discardTransaction();
    }
    log.error(`[performCodePushPromote] ${e}`);
    throw e;
  }
}

export async function performCodePushOtaUpdate(
  napDescriptor: AppVersionDescriptor,
  deploymentName: string,
  miniApps: PackagePath[],
  jsApiImpls: PackagePath[],
  {
    baseComposite,
    codePushIsMandatoryRelease = false,
    codePushRolloutPercentage,
    description,
    disableDuplicateReleaseError = false,
    force = false,
    pathToYarnLock,
    sourceMapOutput,
    targetBinaryVersion,
  }: {
    baseComposite?: PackagePath;
    codePushIsMandatoryRelease?: boolean;
    codePushRolloutPercentage?: number;
    description?: string;
    disableDuplicateReleaseError?: boolean;
    force?: boolean;
    pathToYarnLock?: string;
    sourceMapOutput?: string;
    targetBinaryVersion?: string;
  } = {},
) {
  let cauldron;
  try {
    const codePushSdk = getCodePushSdk();
    cauldron = await getActiveCauldron();
    const plugins = await cauldron.getNativeDependencies(napDescriptor);
    const compositeGenConfig = await cauldron.getCompositeGeneratorConfig(
      napDescriptor,
    );
    baseComposite =
      baseComposite ??
      (compositeGenConfig?.baseComposite &&
        PackagePath.fromString(compositeGenConfig.baseComposite));
    await cauldron.beginTransaction();
    const codePushPlugin = _.find(
      plugins,
      (p) => p.name === 'react-native-code-push',
    );
    if (!codePushPlugin) {
      throw new Error('react-native-code-push plugin is not in native app !');
    }

    const reactNative = _.find(plugins, (p) => p.name === 'react-native');
    if (!codePushPlugin) {
      throw new Error('react-native is not in native app !');
    }

    const tmpWorkingDir = createTmpDir();

    const miniAppsNativeDependenciesVersionAligned =
      await compatibility.areCompatible(miniApps, napDescriptor);

    if (!miniAppsNativeDependenciesVersionAligned && force) {
      log.warn(
        'Native dependencies versions of MiniApps are not aligned but ignoring due to the use of force flag',
      );
    } else if (!miniAppsNativeDependenciesVersionAligned && !force) {
      throw new Error(
        'Native dependencies versions of MiniApps are not aligned. Relaunch the operation with the force flag if you wish to ignore.',
      );
    }

    const latestCodePushedMiniApps = await cauldron.getCodePushMiniApps(
      napDescriptor,
      deploymentName,
    );
    const latestCodePushedJsApiImpls = await cauldron.getCodePushJsApiImpls(
      napDescriptor,
      deploymentName,
    );

    // We need to include, in this CodePush bundle, all the MiniApps and JS API implementations that were part
    // of the previous CodePush. We will override versions of the MiniApps and JS API implementations with
    // the one provided to this function, and keep other ones intact.
    // For example, if previous CodePush bundle was containing first-miniapp@1.0.0 and
    // second-miniapp@1.0.0 and this method is called to CodePush first-miniapp@2.0.0, then
    // the bundle we will push will container first-miniapp@2.0.0 and second-miniapp@1.0.0.
    // If this the first ever CodePush bundle for this specific native application version
    // then the reference miniapp versions are the one from the container.
    let referenceMiniAppsToCodePush = latestCodePushedMiniApps;
    if (
      !referenceMiniAppsToCodePush ||
      referenceMiniAppsToCodePush.length === 0
    ) {
      referenceMiniAppsToCodePush = await cauldron.getContainerMiniApps(
        napDescriptor,
      );
    }

    let referenceJsApiImplsToCodePush = latestCodePushedJsApiImpls;
    if (
      !referenceJsApiImplsToCodePush ||
      referenceJsApiImplsToCodePush.length === 0
    ) {
      referenceJsApiImplsToCodePush = await cauldron.getContainerJsApiImpls(
        napDescriptor,
      );
    }

    const miniAppsToBeCodePushedUnordered: PackagePath[] = _.unionBy(
      miniApps,
      referenceMiniAppsToCodePush,
      (x) => x.basePath,
    );

    // Reorder MiniApps to preserve current order set in Cauldron
    const miniAppsToBeCodePushed: PackagePath[] = orderPackagePaths(
      referenceMiniAppsToCodePush,
      miniAppsToBeCodePushedUnordered,
    );

    const jsApiImplsToBeCodePushedUnordered = _.unionBy(
      jsApiImpls,
      referenceJsApiImplsToCodePush,
      (x) => x.basePath,
    );

    // Reorder JS API implementations to preserve current order set in Cauldron
    const jsApiImplsToBeCodePushed: PackagePath[] = orderPackagePaths(
      referenceJsApiImplsToCodePush,
      jsApiImplsToBeCodePushedUnordered,
    );

    const pathsToMiniAppsToBeCodePushed = _.map(miniAppsToBeCodePushed, (m) =>
      PackagePath.fromString(m.toString()),
    );
    const pathToJsApiImplsToBeCodePushed = _.map(
      jsApiImplsToBeCodePushed,
      (j) => PackagePath.fromString(j.toString()),
    );

    const codePushConfig = await cauldron.getCodePushConfig(napDescriptor);
    if (codePushConfig && codePushConfig.bypassYarnLock) {
      pathToYarnLock = undefined;
      log.debug(
        'Bypassing yarn.lock usage as bypassYarnLock flag is set in Cauldron config',
      );
    }

    const composite = await kax.task('Generating composite module').run(
      GeneratedComposite.generate({
        baseComposite,
        jsApiImplDependencies: pathToJsApiImplsToBeCodePushed,
        miniApps: pathsToMiniAppsToBeCodePushed,
        outDir: tmpWorkingDir,
        pathToYarnLock,
        resolutions: compositeGenConfig && compositeGenConfig.resolutions,
      }),
    );

    const bundleOutputDirectory = path.join(tmpWorkingDir, 'bundleOut');
    shell.mkdir('-p', bundleOutputDirectory);
    const platform = napDescriptor.platform || '';
    const bundleOutputPath =
      platform === 'android'
        ? path.join(bundleOutputDirectory, 'index.android.bundle')
        : path.join(bundleOutputDirectory, 'MiniApp.jsbundle');

    sourceMapOutput = sourceMapOutput || path.join(createTmpDir(), 'index.map');
    const entryFile = fs.existsSync(
      path.join(tmpWorkingDir, `index.${platform}.js`),
    )
      ? `index.${platform}.js`
      : 'index.js';
    const bundlingResult = await kax
      .task('Generating composite bundle for miniapps')
      .run(
        reactnative.bundle({
          assetsDest: bundleOutputDirectory,
          bundleOutput: bundleOutputPath,
          dev: false,
          entryFile,
          platform,
          resetCache: true,
          sourceMapOutput,
          workingDir: tmpWorkingDir,
        }),
      );

    if (platform === 'android') {
      const conf = await cauldron.getContainerGeneratorConfig(napDescriptor);
      const isHermesEnabled = conf?.androidConfig?.jsEngine === 'hermes';
      if (isHermesEnabled) {
        const hermesVersion =
          conf.androidConfig.hermesVersion ??
          android.getDefaultHermesVersion(reactNative?.version!);
        const hermesCli = await kax
          .task(`Installing hermes-engine@${hermesVersion}`)
          .run(HermesCli.fromVersion(hermesVersion));
        const res = await kax
          .task('Compiling JS bundle to Hermes bytecode')
          .run(
            hermesCli.compileReleaseBundle({
              bundleSourceMapPath: sourceMapOutput,
              compositePath: tmpWorkingDir,
              jsBundlePath: bundleOutputPath,
            }),
          );
        bundlingResult.isHermesBundle = true;
      }
    }

    const appName = await getCodePushAppName(napDescriptor);

    targetBinaryVersion =
      targetBinaryVersion ||
      (await buildCodePushTargetBinaryVersion(napDescriptor, deploymentName));

    log.info(`Target Binary version : ${targetBinaryVersion}`);
    description = description || '';
    const codePushResponse: CodePushPackage | void = await kax
      .task('Releasing bundle through CodePush')
      .run(
        codePushSdk.releaseReact(
          appName,
          deploymentName,
          bundleOutputDirectory,
          targetBinaryVersion,
          {
            description,
            isMandatory: codePushIsMandatoryRelease,
            rollout: codePushRolloutPercentage,
          },
          disableDuplicateReleaseError,
        ),
      );
    if (codePushResponse) {
      await cauldron.addCodePushEntry(
        napDescriptor,
        {
          appVersion: codePushResponse.appVersion,
          deploymentName,
          description: codePushResponse.description,
          isMandatory: codePushResponse.isMandatory,
          label: codePushResponse.label,
          releaseMethod: codePushResponse.releaseMethod,
          releasedBy: codePushResponse.releasedBy,
          rollout: codePushResponse.rollout,
          size: codePushResponse.size,
        },
        miniAppsToBeCodePushed,
        jsApiImplsToBeCodePushed,
      );

      const pathToNewYarnLock = path.join(tmpWorkingDir, 'yarn.lock');
      if (await fs.pathExists(pathToNewYarnLock)) {
        await kax
          .task('Adding yarn.lock to Cauldron')
          .run(
            cauldron.addOrUpdateYarnLock(
              napDescriptor,
              deploymentName,
              pathToNewYarnLock,
            ),
          );
      }

      // Upload source map if a source map store server is configured
      const sourcemapStoreConfig = await cauldron.getSourceMapStoreConfig();
      if (sourcemapStoreConfig) {
        try {
          const sdk = new SourceMapStoreSdk(sourcemapStoreConfig.url);
          await kax
            .task(
              `Uploading source map to source map store [${sourcemapStoreConfig.url}]`,
            )
            .run(
              sdk.uploadCodePushSourceMap({
                deploymentName,
                descriptor: napDescriptor,
                label: codePushResponse.label!,
                sourceMapPath: sourceMapOutput,
              }),
            );
        } catch (e) {
          log.error(`Source map upload failed : ${e}`);
        }
      }

      // Upload source map to bugsnag if configured
      const bugsnagConfig = await cauldron.getBugsnagConfig(napDescriptor);
      if (bugsnagConfig) {
        try {
          const { apiKey } = bugsnagConfig;
          const [minifiedFile, minifiedUrl, projectRoot, sourceMap] = [
            await fs.realpath(bundleOutputPath),
            path.basename(bundleOutputPath),
            await fs.realpath(path.join(composite.path, 'node_modules')),
            await fs.realpath(sourceMapOutput),
          ];
          const compositeMiniAppsPackages =
            await composite.getMiniAppsPackages();
          await bugsnagUpload({
            apiKey,
            minifiedFile,
            minifiedUrl,
            projectRoot,
            sourceMap,
            uploadNodeModules: false,
            uploadSources: false,
          });
        } catch (e) {
          log.error(`Bugsnag upload failed : ${e}`);
        }
      }

      await cauldron.commitTransaction(
        `CodePush release for ${napDescriptor.toString()} ${deploymentName}`,
      );
    }
  } catch (e) {
    if (cauldron) {
      await cauldron.discardTransaction();
    }
    log.error(`performCodePushOtaUpdate ${e}`);

    throw e;
  }
}

export async function buildCodePushTargetBinaryVersion(
  napDescriptor: AppVersionDescriptor,
  deploymentName: string,
) {
  if (!napDescriptor.version) {
    throw new Error(
      `Native application descriptor ${napDescriptor} does not contain a version !`,
    );
  }
  let targetBinaryVersion = napDescriptor.version;
  const cauldron = await getActiveCauldron();
  const codePushConfig = await cauldron.getCodePushConfig(napDescriptor);
  if (codePushConfig) {
    const { versionModifiers, trimZeroPatchDigit } = codePushConfig;
    targetBinaryVersion = versionModifiers
      ? applyVersionModifiers({
          deploymentName,
          targetBinaryVersion,
          versionModifiers,
        })
      : targetBinaryVersion;
    targetBinaryVersion =
      trimZeroPatchDigit &&
      semver.valid(targetBinaryVersion) &&
      semver.patch(targetBinaryVersion) === 0
        ? removeZeroPatchDigit({
            targetBinaryVersion,
          })
        : targetBinaryVersion;
  }

  return targetBinaryVersion;
}

export function applyVersionModifiers({
  deploymentName,
  targetBinaryVersion,
  versionModifiers,
}: {
  deploymentName: string;
  targetBinaryVersion: string;
  versionModifiers: { deploymentName: string; modifier: string }[];
}): string {
  const versionModifier = _.find(
    versionModifiers,
    (m) => m.deploymentName === deploymentName,
  );
  return versionModifier
    ? targetBinaryVersion.replace(/(.+)/, versionModifier.modifier)
    : targetBinaryVersion;
}

export function removeZeroPatchDigit({
  targetBinaryVersion,
}: {
  targetBinaryVersion: string;
}) {
  return semver.prerelease(targetBinaryVersion)
    ? `${semver.major(targetBinaryVersion)}.${semver.minor(
        targetBinaryVersion,
      )}-${semver.prerelease(targetBinaryVersion)!.join('.')}`
    : `${semver.major(targetBinaryVersion)}.${semver.minor(
        targetBinaryVersion,
      )}`;
}

export async function getCodePushAppName(
  napDescriptor: AppVersionDescriptor,
): Promise<string> {
  const cauldron = await getActiveCauldron();
  const codePushConfig = await cauldron.getCodePushConfig(napDescriptor);
  const cauldronCodepushAppName: CauldronNativeApp[] = (
    (await cauldron.getAllNativeApps()) as CauldronNativeApp[]
  )
    .find((app) => app.name === napDescriptor.name)
    ?.platforms.find((platform) => platform.name === napDescriptor.platform)
    ?.config?.codePush?.appName;
  return cauldronCodepushAppName
    ? cauldronCodepushAppName
    : codePushConfig?.appName ??
        `${napDescriptor.name}${
          napDescriptor.platform === 'ios' ? 'Ios' : 'Android'
        }`;
}

// Reorder packages to preserve current order as set in Cauldron
export function orderPackagePaths(
  referencePackagePaths: PackagePath[],
  packagePaths: PackagePath[],
): PackagePath[] {
  const packages: PackagePath[] = [];
  for (const m of referencePackagePaths) {
    const x = _.find(packagePaths, (p) => p.basePath === m.basePath) ?? m;
    packages.push(x);
  }
  // Add all new packages if any
  const newPackages: PackagePath[] = _.differenceBy(
    packagePaths,
    referencePackagePaths,
    (p) => p.basePath,
  );
  packages.push(...newPackages);
  return packages;
}
